import email.parser
import re
from typing import Dict, Set, Callable, BinaryIO, Iterable

from packaging.version import Version
from packaging.requirements import Requirement
from packaging.utils import canonicalize_name
from packaging.specifiers import SpecifierSet
from packaging.markers import Marker, Variable as MarkerVariable
import tomli


METADATA_VERSION_11 = Version("1.1")
METADATA_VERSION_12 = Version("1.2")
METADATA_VERSION_21 = Version("2.1")


class MetadataParser:
    """
    MetadataParser is used to incrementally parse metadata sources from a Python
    package (which could be a wheel archive, a source archive, or just a
    directory). The major use of the class is to resolve dependencies of the
    package.

    The class supports several sources:
    - Metadata files compliant with Python's "core metadata specifications"
      (https://packaging.python.org/en/latest/specifications/core-metadata/),
      e.g. METADATA and PKG-INFO files.
    - The pyproject.toml file, as per PEP 621 (https://peps.python.org/pep-0621).
    - Setuptools-specific requires.txt and setup-requires.txt files
      (https://setuptools.pypa.io/en/latest/deprecated/python_eggs.html#requires-txt)

    Attributes:
    -----------
    source_path : str
        The name of the package archive that the object was
        created to parse.
    name : str
        The name of the package, in canonicalized form
    version : packaging.version.Version
        The version of the package
    python_requirement : packaging.specifiers.SpecifierSet
        A specification of the Python versions supported by the package.
    extras_provided : Set[str]
        Extras provided by the package.
    core_dependencies: Set[packaging.requirements.Requirement]
        Core dependencies of the package.
    optional_dependencies: Dict[str, Set[packaging.requirements.Requirement]]
        Optional dependencies of the package. A dictionary whose keys are
        either names of extras (from extras_provided) or environment marker
        constraints (e.g. :python_version<2.7).
    build_dependencies : Set[packaging.requirements.Requirement]
        Dependencies required to build the package.
    """

    def __init__(self, source_path: str):
        """
        Object constructor.

        Parameters:
        -----------
        filename : str
            The name of the package archive that the object
            was created to parse.

        """

        self.source_path: str = source_path
        self.name: str = None
        self.version: Version = None
        self.python_requirement: SpecifierSet = None
        self.extras_provided: Set[str] = set()
        self.core_dependencies: Set[Requirement] = set()
        self.optional_dependencies: Dict[str, Set[Requirement]] = {}
        self.build_dependencies: Set[Requirement] = set()

    def parse(
        self,
        opener: Callable[[str], BinaryIO],
        filename: str,
    ):
        """
        Parses a file, gathering whatever metadata can be gathered from it. Any
        file can be provided to the method, irrelevant files are simply ignored.

        Parameters
        ----------
        opener : Callable[[str], BinaryIO]
            A function that can be used to open the file. The function takes one
            parameter, which is the file name, and returns a file object opened
            in binary mode.
        filename : str
            The file name that will be provided to the opener function. This can
            either be relative or absolute, so long as the opener function can
            open it. The name is important, as the method uses it to determine
            the kind of file it is.
        """

        parse_func = None
        main_metadata_file = False

        if re.search(r"\.whl$", self.source_path):
            if re.fullmatch(r"[^/]+\.dist-info/METADATA", filename):
                parse_func = self._parse_metadata_file
                main_metadata_file = True
        elif re.search(r"\.zip$", self.source_path):
            if re.fullmatch(r"([^/]+/)?PKG-INFO", filename):
                parse_func = self._parse_metadata_file
                main_metadata_file = True
        elif re.search(r"\.tar\.gz$", self.source_path):
            if re.fullmatch(r"[^/]+/PKG-INFO", filename):
                parse_func = self._parse_metadata_file
                main_metadata_file = True
            elif re.fullmatch(r"[^/]+(/[^/]+)?\.egg-info/(setup_)?requires.txt", filename):
                parse_func = self._parse_requirestxt
            elif re.fullmatch(r"[^/]+/pyproject.toml", filename):
                parse_func = self._parse_pyproject

        if parse_func:
            with opener(filename) as fp:
                if main_metadata_file:
                    self._metadata_file = fp.read()
                    fp.seek(0)
                parse_func(fp)

    def seen_metadata_file(self) -> bool:
        """
        Returns a boolean value if the archive's main METADATA file has already
        been read.
        """
        return hasattr(self, "_metadata_file")

    def write_metadata_file(self, target: str):
        """
        Writes the archive's main METADATA file to the target filepath, provided
        such a file was already read. The contents of the target file will be
        clobbered, if already existing. If the main metadata file has not been
        read yet, an exception will be raised.
        """
        if not hasattr(self, "_metadata_file"):
            raise Exception("Main METADATA file has not been read yet")

        with open(target, "wb") as out:
            out.write(self._metadata_file)

    def dependencies(
        self,
        extras: Set[str] = set(),
        envs: Iterable[Dict] = []
    ) -> Set[Requirement]:
        """
        Resolves the dependencies of the package, returning a set of
        requirements. Only requirements that are relevant to the provided extras
        and environments are returned.

        Parameters
        ----------
        extras : Set[str] = set()
            A set of extras that the package was required with. For example, if
            the instance of this class is used to parse the metadata of the
            package "pymongo", and the requirement string for that package was
            "pymongo[snappy,zstd]", then the set of extras will be (snappy, zstd).
        envs: Iterable[Dict] = []
            The list of environments for which Morgan is downloading package
            distributions. These are simple dictionaries whose keys match those
            defined by the "Environment Markers" section of PEP 508.

        Returns
        -------
        A set of packaging.requirements.Requirement objects.
        """

        deps = set()
        deps |= self.core_dependencies
        deps |= self.build_dependencies

        for extra in self.optional_dependencies:
            if ":" in extra:
                # this dependency includes a set of environment marker
                # specifications
                orig = extra
                (extra, spec) = extra.split(":")
                if extra and extra not in extras:
                    continue
                marker = Marker(spec)
                for env in envs:
                    if marker.evaluate(env):
                        deps |= self.optional_dependencies[orig]
                        break
            elif extra in extras:
                deps |= self.optional_dependencies[extra]

        irrelevant_deps = set()
        for dep in deps:
            relevant = True
            if dep.marker:
                relevant = False
                for env in envs:
                    env["extra"] = ",".join(extras)
                    if dep.marker.evaluate(env):
                        relevant = True
                        break

            if not relevant:
                irrelevant_deps.add(dep)
                continue

        deps -= irrelevant_deps

        return deps

    def _add_core_requirements(self, reqs):
        self.core_dependencies |= set([Requirement(dep) for dep in reqs])

    def _add_optional_requirements(self, extra, reqs):
        if extra not in self.optional_dependencies:
            self.optional_dependencies[extra] = set()
        self.optional_dependencies[extra] |= set(
            [Requirement(dep) for dep in reqs])

    def _parse_pyproject(self, fp):
        data = tomli.load(fp)
        project = data.get("project")

        if project is not None:
            (name, version) = (project.get("name"), project.get("version"))

            if name is not None:
                self.name = canonicalize_name(name)

            if version is not None:
                self.version = Version(version)

            if "requires-python" in project:
                self.python_requirement = SpecifierSet(
                    project["requires-python"])

            if "dependencies" in project:
                self._add_core_requirements(project["dependencies"])

            if "optional-dependencies" in project:
                for extra in project["optional-dependencies"]:
                    self._add_optional_requirements(
                        extra, project["optional-dependencies"][extra])

        build_system = data.get("build-system")
        if build_system is not None and "requires" in build_system:
            self.build_dependencies |= set(
                [Requirement(req) for req in build_system["requires"]])

    def _parse_metadata_file(self, fp):
        data = email.parser.BytesParser().parse(fp, True)

        (name, version, metadata_version) = (data.get("Name"),
                                             data.get("Version"),
                                             data.get("Metadata-Version"))
        if metadata_version is None:
            return

        metadata_version = Version(metadata_version)

        if name is not None:
            self.name = canonicalize_name(name)
        if version is not None:
            self.version = Version(version)

        if metadata_version >= METADATA_VERSION_21:
            self._parse_metadata_21(data)
        if metadata_version >= METADATA_VERSION_12:
            self._parse_metadata_12(data)
        elif metadata_version == METADATA_VERSION_11:
            self._parse_metadata_11(data)

    def _parse_metadata_21(self, data):
        provides_extra = data.get_all("Provides-Extra")
        if provides_extra is not None:
            self.extras_provided |= set(provides_extra)

    def _parse_metadata_12(self, data):
        requires_python = data.get("Requires-Python")
        if requires_python is not None:
            self.python_requirement = SpecifierSet(requires_python)

        requires_dist = data.get_all("Requires-Dist")
        if requires_dist is not None:
            for requirement_str in requires_dist:
                req = Requirement(requirement_str)
                extra = None
                if req.marker is not None:
                    for marker in req.marker._markers:
                        if isinstance(marker[0], MarkerVariable) and \
                                marker[0].value == "extra":
                            extra = marker[2].value
                            break

                if extra:
                    if extra not in self.optional_dependencies:
                        self.optional_dependencies[extra] = set()
                    self.optional_dependencies[extra].add(req)
                else:
                    self.core_dependencies.add(req)

    def _parse_metadata_11(self, data):
        requires = data.get_all("Requires")
        if requires is not None:
            for requirement_str in requires:
                self.core_dependencies.add(Requirement(requirement_str))

    def _parse_requirestxt(self, fp):
        section = None
        content = []
        for line in fp.readlines():
            line = line.strip().decode("UTF-8")
            if line.startswith("["):
                if line.endswith("]"):
                    if section or content:
                        if section:
                            self._add_optional_requirements(section, content)
                        elif fp.name.endswith("setup_requires.txt"):
                            self._add_build_requirements(content)
                        else:
                            self._add_core_requirements(content)
                    section = line[1:-1]
                    content = []
                else:
                    raise ValueError("Invalid section heading", line)
            elif line:
                content.append(line)

        if section:
            self._add_optional_requirements(section, content)
        elif fp.name.endswith("setup_requires.txt"):
            self._add_build_requirements(content)
        else:
            self._add_core_requirements(content)



import argparse
from collections import OrderedDict
import configparser
import os
import platform
import sys

from packaging.version import Version

if Version(platform.python_version()) < Version('3.8'):
    import importlib_metadata as metadata
else:
    import importlib.metadata as metadata


def generate_env(name: str = "local"):
    """
    Generate a configuration block for the local client environment. This is
    an implementation of the PEP 345 specification of "Environment Markers".
    Resulting block is printed to standard output, and can either be copied to
    the configuration file, or piped to it using shell redirection (e.g. `>>`).
    """

    config = configparser.ConfigParser()
    config["env.{}".format(name)] = {
        'os_name': os.name,
        'sys_platform': sys.platform,
        'platform_machine': platform.machine(),
        'platform_python_implementation': platform.python_implementation(),
        'platform_system': platform.system(),
        'python_version': '.'.join(platform.python_version_tuple()[:2]),
        'python_full_version': platform.python_version(),
        'implementation_name': sys.implementation.name,
    }
    config.write(sys.stdout)


def generate_reqs(mode: str = ">="):
    """
    Generate a requirements configuration block from current environment.

    The requirements block is printed to standard output,
    and can either be copied to the configuration file, or piped to it
    using shell redirection (e.g. `>>`).

    Args:
        mode (str, optional):
            Mode to use for versioning. Use "==" for exact versioning,
            ">=" for minimum versioning, or "<=" for maximum versioning.
            Defaults to ">=".
    """
    requirements = {dist.metadata["Name"].lower(): f"{mode}{dist.version}"
                    for dist in metadata.distributions()}
    config = configparser.ConfigParser()
    config["requirements"] = OrderedDict(sorted(requirements.items()))
    config.write(sys.stdout)


def add_arguments(parser: argparse.ArgumentParser):
    """
    Adds command line options specific to this script to an argument parser.
    """

    parser.add_argument(
        '-e', '--env',
        dest='env',
        help='Name of environment to configure'
    )

    parser.add_argument(
        '-m', '--mode',
        dest='mode',
        choices=['>=', '==', '<='],
        default=">=",
        help='Versioning mode for requirements',
    )



import pytest

from morgan import server


@pytest.mark.parametrize(
    "accept_option, exp_dict",
    [
        (
            server.GENL_HTML_TYPE,
            {"mime": server.GENL_HTML_TYPE, "priority": 0}
        ),
        (
            "{};q=1".format(server.GENL_HTML_TYPE),
            {"mime": server.GENL_HTML_TYPE, "priority": 1}
        ),
        (
            "{};q=0.9".format(server.GENL_HTML_TYPE),
            {"mime": server.GENL_HTML_TYPE, "priority": 0.9}
        ),
        (
            "{} ; q=0.9".format(server.GENL_HTML_TYPE),
            {"mime": server.GENL_HTML_TYPE, "priority": 0.9}
        ),
        (
            "{};q=0.9&charset=UTF-8".format(server.GENL_HTML_TYPE),
            {"mime": server.GENL_HTML_TYPE, "priority": 0.9}
        ),
        (
            "{};charset=UTF-8&q=0.9".format(server.GENL_HTML_TYPE),
            {"mime": server.GENL_HTML_TYPE, "priority": 0.9}
        ),
    ],
)
def test_parse_accept_option(accept_option, exp_dict):
    got_dict = server.parse_accept_option(accept_option)
    assert got_dict == exp_dict


@pytest.mark.parametrize(
    "accept_header, exp_mime",
    [
        (
            None,
            server.PYPI_HTML_TYPE_V1
        ),
        (
            "text/xml",
            None,
        ),
        (
            server.PYPI_JSON_TYPE_V1,
            server.PYPI_JSON_TYPE_V1,
        ),
        (
            "{};q=0.5, {}; q=1".format(
                server.PYPI_JSON_TYPE_V1, server.GENL_HTML_TYPE),
            server.GENL_HTML_TYPE
        ),
        (
            "*/*",
            server.PYPI_HTML_TYPE_V1,
        ),
        (
            "text/xml;q=1,*/*; q=0.5",
            server.PYPI_HTML_TYPE_V1,
        ),
    ],
)
def test_parse_accept_header(accept_header, exp_mime):
    got_mime = server.parse_accept_header(accept_header)
    assert got_mime == exp_mime



import argparse
import configparser
import hashlib
import json
import os
import os.path
import re
import tarfile
import traceback
from typing import Iterable, Tuple, Dict
import urllib.parse
import urllib.request
import zipfile

import packaging.requirements
import packaging.version
import packaging.utils
import packaging.tags
import packaging.specifiers

from morgan import server, configurator, metadata
from morgan.__about__ import __version__

PYPI_ADDRESS = "https://pypi.org/simple/"
PREFERRED_HASH_ALG = "sha256"


class Mirrorer:
    """
    Mirrorer is a class that implements the mirroring capabilities of Morgan.
    A class is used to maintain state, as the mirrorer needs to keep track of
    packages it already processed in the (very common) case that it encounters
    them again as dependencies.
    """

    def __init__(self, index_path: str):
        """
        The constructor only needs to path to the package index.
        """

        # load the configuration from the index_path, and parse the environments
        # into representations that are easier for the mirrorer to work with
        self.index_path = index_path
        self.config = configparser.ConfigParser()
        self.config.read(os.path.join(self.index_path, "morgan.ini"))
        self.envs = {}
        self._supported_pyversions = []
        self._supported_platforms = []
        for key in self.config:
            m = re.fullmatch(r"env\.(.+)", key)
            if m:
                env = self.config[key]
                env['platform_release'] = ''
                env['platform_version'] = ''
                env['implementation_version'] = ''
                env['extra'] = ''
                self.envs[m.group(1)] = dict(env)
                self._supported_pyversions.append(env["python_version"])
                self._supported_platforms.append(
                    re.compile(r".*" +
                               env["sys_platform"] +
                               r".*" +
                               env["platform_machine"]))

        self._processed_pkgs = {}

    def mirror(self, requirement_string: str):
        """
        Mirror a package according to a PEP 508-compliant requirement string.
        """

        requirement = parse_requirement(requirement_string)

        deps = self._mirror(requirement)
        if deps is None:
            return

        while len(deps) > 0:
            next_deps = {}
            for dep in deps:
                more_deps = self._mirror(
                    deps[dep]["requirement"],
                    required_by=deps[dep]["required_by"],
                )
                if more_deps:
                    next_deps.update(more_deps)
            deps = next_deps.copy()

    def copy_server(self):
        """
        Copy the server script to the package index. This method will first
        attempt to find the server file directly, and if that fails, it will
        use the inspect module to get the source code.
        """

        print("Copying server script")
        thispath = os.path.realpath(__file__)
        serverpath = os.path.join(os.path.dirname(thispath), "server.py")
        outpath = os.path.join(self.index_path, "server.py")
        if os.path.exists(serverpath):
            with open(serverpath, "rb") as inp, open(outpath, "wb") as out:
                out.write(inp.read())
        else:
            import inspect
            with open(outpath, "w") as out:
                out.write(inspect.getsource(server))

    def _mirror(
        self,
        requirement: packaging.requirements.Requirement,
        required_by: packaging.requirements.Requirement = None,
    ) -> dict:
        if requirement.name in self._processed_pkgs:
            return None

        if required_by:
            print("[{}]: {}".format(required_by, requirement))
        else:
            print("{}".format(requirement))

        data: dict = None

        # get information about this package from the Simple API in JSON
        # format as per PEP 691
        request = urllib.request.Request(
            "{}{}/".format(PYPI_ADDRESS, requirement.name),
            headers={
                'Accept': 'application/vnd.pypi.simple.v1+json',
            },
        )

        with urllib.request.urlopen(request) as response:
            data = json.load(response)

        if data["meta"]["api-version"] != "1.0":
            raise Exception(
                "Unsupported metadata version {}, only support 1.0".format(
                    data["meta"]["api-version"]))

        files = data["files"]
        if files is None or not isinstance(files, list):
            raise Exception(
                "Expected response to contain a list of 'files'")

        # filter and enrich files
        files = self._filter_files(requirement, files)
        if files is None:
            if required_by is None:
                raise Exception("No files match requirement")
            else:
                # this is a dependency, assume the dependency is not relevant
                # for any of our environments and don't return an error
                return None

        # download all files
        depdict = {}
        for file in files:
            try:
                file_deps = self._process_file(requirement, file)
                if file_deps:
                    depdict.update(file_deps)
            except Exception:
                print("\tFailed processing file {}, skipping it".format(
                    file["filename"]))
                traceback.print_exc()
                continue

        self._processed_pkgs[requirement.name] = True

        return depdict

    def _filter_files(
        self,
        requirement: packaging.requirements.Requirement,
        files: Iterable[dict],
    ) -> Iterable[dict]:
        # remove files with unsupported extensions
        files = list(filter(lambda file: re.search(
            r"\.(whl|zip|tar.gz)$", file["filename"]), files))

        # parse versions and platform tags for each file
        for file in files:
            try:
                if re.search(r"\.whl$", file["filename"]):
                    _, file["version"], ___, file["tags"] = \
                        packaging.utils.parse_wheel_filename(
                            file["filename"])
                    file["is_wheel"] = True
                elif re.search(r"\.(tar\.gz|zip)$", file["filename"]):
                    _, file["version"] = packaging.utils.parse_sdist_filename(
                        file["filename"])
                    file["is_wheel"] = False
                    file["tags"] = None
            except packaging.version.InvalidVersion:
                # ignore files with invalid version, PyPI no longer allows
                # packages with special versioning schemes, and we assume we
                # can ignore such files
                continue
            except Exception:
                print("\tSkipping file {}, exception caught".format(
                    file["filename"]))
                traceback.print_exc()
                continue

        # sort all files by version in reverse order, and ignore yanked files
        files = list(filter(lambda file: "version" in file and
                            not file.get("yanked", False), files))
        files.sort(key=lambda file: file["version"], reverse=True)

        # keep only files of the latest version that satisfies the
        # requirement (if requirement doesn't have any version specifiers,
        # take latest available version)
        if requirement.specifier is not None:
            files = list(filter(
                lambda file: requirement.specifier.contains(
                    file["version"]),
                files))

        if len(files) == 0:
            print("Skipping {}, no version matches requirement".format(
                requirement.name))
            return None

        latest_version = files[0]["version"]
        files = list(filter(
            lambda file: file["version"] == latest_version, files))

        # now we only have files from the latest version that satisfies the
        # requirement. we need to filter out files that do not match any of
        # our environments
        files = list(filter(
            lambda file: self._should_download(file), files))

        return files

    def _should_download(self, fileinfo: dict) -> bool:
        if fileinfo.get("requires-python", None):
            spec_set = packaging.specifiers.SpecifierSet(
                fileinfo["requires-python"])
            supported = False
            for supported_python in self._supported_pyversions:
                if spec_set.contains(supported_python):
                    supported = True
                    break
            if not supported:
                return False

        if fileinfo.get("tags", None):
            for tag in fileinfo["tags"]:
                (intrp_name, intrp_ver) = parse_interpreter(tag.interpreter)
                if intrp_name not in ("py", "cp"):
                    continue
                if (intrp_ver and
                        intrp_ver != "3" and
                        intrp_ver not in self._supported_pyversions):
                    continue

                if tag.platform == "any":
                    return True
                else:
                    for platformre in self._supported_platforms:
                        if platformre.fullmatch(tag.platform):
                            return True
            return False

        return True

    def _process_file(
        self,
        requirement: packaging.requirements.Requirement,
        fileinfo: dict,
    ) -> Dict[str, packaging.requirements.Requirement]:
        filepath = os.path.join(
            self.index_path, requirement.name, fileinfo["filename"])
        hashalg = PREFERRED_HASH_ALG\
            if PREFERRED_HASH_ALG in fileinfo["hashes"]\
            else fileinfo["hashes"].keys()[0]

        self._download_file(fileinfo, filepath, hashalg)

        md = self._extract_metadata(
            filepath, requirement.name, fileinfo["version"])

        deps = md.dependencies(requirement.extras, self.envs.values())
        if deps is None:
            return None

        depdict = {}
        for dep in deps:
            dep.name = packaging.utils.canonicalize_name(dep.name)
            depdict[dep.name] = {
                'requirement': dep,
                'required_by': requirement,
            }
        return depdict

    def _download_file(
        self,
        fileinfo: dict,
        target: str,
        hashalg: str,
    ) -> bool:
        exphash = fileinfo["hashes"][hashalg]

        os.makedirs(os.path.dirname(target), exist_ok=True)

        # if target already exists, verify its hash and only download if
        # there's a mismatch
        if os.path.exists(target):
            truehash = self._hash_file(target, hashalg)
            if truehash == exphash:
                return True

        print("\t{}...".format(fileinfo["url"]), end=" ")
        with urllib.request.urlopen(fileinfo["url"]) as inp,\
                open(target, "wb") as out:
            out.write(inp.read())
        print("done")

        truehash = self._hash_file(target, hashalg)
        if truehash != exphash:
            raise Exception(
                "Digest mismatch for {}".format(fileinfo["filename"]))

        return True

    def _hash_file(self, filepath: str, hashalg: str) -> str:
        contents = None
        with open(filepath, "rb") as fh:
            # verify downloaded file has same hash
            contents = fh.read()

        truehash = hashlib.new(hashalg)
        truehash.update(contents)

        with open("{}.hash".format(filepath), "w") as out:
            out.write("{}={}".format(hashalg, truehash.hexdigest()))

        return truehash.hexdigest()

    def _extract_metadata(
        self,
        filepath: str,
        package: str,
        version: packaging.version.Version,
    ) -> metadata.MetadataParser:
        md = metadata.MetadataParser(filepath)

        archive = None
        members = None
        opener = None

        if re.search(r"\.(whl|zip)$", filepath):
            archive = zipfile.ZipFile(filepath)
            members = [member.filename for member in archive.infolist()]
            opener = archive.open
        elif re.search(r"\.tar.gz$", filepath):
            archive = tarfile.open(filepath)
            members = [member.name for member in archive.getmembers()]
            opener = archive.extractfile
        else:
            raise Exception("Unexpected distribution file {}".format(filepath))

        for member in members:
            try:
                md.parse(opener, member)
            except Exception as e:
                print("Failed parsing member {} of {}: {}".format(
                    member, filepath, e))

        if md.seen_metadata_file():
            md.write_metadata_file("{}.metadata".format(filepath))

        archive.close()

        return md


def parse_interpreter(inp: str) -> Tuple[str, str]:
    """
    Parse interpreter tags in the name of a binary wheel file. Returns a tuple
    of interpreter name and optional version, which will either be <major> or
    <major>.<minor>.
    """

    m = re.fullmatch(r"^([^\d]+)(?:(\d)(?:[._])?(\d+)?)$", inp)
    if m is None:
        return (inp, None)

    intr = m.group(1)
    version = None
    if m.lastindex > 1:
        version = m.group(2)
        if m.lastindex > 2:
            version = "{}.{}".format(version, m.group(3))

    return (intr, version)


def parse_requirement(req_string: str) -> packaging.requirements.Requirement:
    """
    Parse a requirement string into a packaging.requirements.Requirement object.
    Also canonicalizes (or "normalizes") the name of the package.
    """

    req = packaging.requirements.Requirement(req_string)
    req.name = packaging.utils.canonicalize_name(req.name)
    return req


def mirror(index_path: str):
    """
    Run the mirror on the package index in the provided path, and based on the
    morgan.ini configuration file in the index. Copies the server script to the
    index at the end of the process. This function can safely be called multiple
    times on the same index path, files are only downloaded if necessary.
    """

    m = Mirrorer(index_path)
    for package in m.config["requirements"]:
        m.mirror("{}{}".format(package, m.config["requirements"][package]))
    m.copy_server()


def main():
    """
    Executes the command line interface of Morgan. Use -h for a full list of
    flags, options and arguments.
    """

    parser = argparse.ArgumentParser(
        description='Morgan: PyPI Mirror for Restricted Environments')

    parser.add_argument(
        '-i', '--index-path',
        dest='index_path',
        default=os.getcwd(),
        help='Path to the package index')

    server.add_arguments(parser)
    configurator.add_arguments(parser)

    parser.add_argument(
        "command",
        choices=["generate_env", "generate_reqs", "mirror", "serve",
            "copy_server", "version"],
        help="Command to execute")

    args = parser.parse_args()

    if args.command == "serve":
        server.run(args.index_path, args.host, args.port)
    elif args.command == "generate_env":
        configurator.generate_env(args.env)
    elif args.command == "generate_reqs":
        configurator.generate_reqs(args.mode)
    elif args.command == "mirror":
        mirror(args.index_path)
    elif args.command == "copy_server":
        Mirrorer(args.index_path).copy_server()
    elif args.command == "version":
        print("Morgan v{}".format(__version__))

if __name__ == "__main__":
    main()



import argparse
import html
import http.server
import json
import os
import pathlib
import re
import urllib.parse

PYPI_JSON_TYPE_V1 = 'application/vnd.pypi.simple.v1+json'
PYPI_JSON_TYPE_LT = 'application/vnd.pypi.simple.latest+json'
PYPI_HTML_TYPE_V1 = 'application/vnd.pypi.simple.v1+html'
GENL_HTML_TYPE = 'text/html'

project_re = re.compile(r"/([^/]+)/")
file_re = re.compile(r"/([^/]+)/([^/]+)")
index_path = os.getcwd()


class RequestHandler(http.server.BaseHTTPRequestHandler):
    """
    The request handler class for the server. Extends
    http.server.BaseHTTPRequestHandler. Implements the Simple Repository API
    (PEP 503) with HTML output, or JSON output as per PEP 691. Also serves
    metadata files for packages as per PEP 658. Only GET requests are currently
    supported.
    """

    def do_GET(self):
        url = urllib.parse.urlsplit(self.path)

        ct = parse_accept_header(self.headers.get("Accept"))
        if ct is None:
            self.send_response(406)
            self.send_header("Content-Type", "text/plain")
            self.end_headers()
            self.wfile.write(
                b"The server cannot generate a response " +
                b"in any of the requested MIME types")
            return

        if url[2] in ["", "/"]:
            self._serve_project_listing(ct)
            return

        m = project_re.fullmatch(url[2])
        if m and m.group(1) != "":
            self._serve_project(ct, m.group(1))
            return

        m = file_re.fullmatch(url[2])
        if m and m.group(1) != "" and m.group(2) != "":
            self._serve_file(m.group(1), m.group(2))
            return

        self._serve_notfound()

    def _serve_notfound(self, msg: str = None):
        self.send_response(404)
        self.send_header("Content-Type", "text/plain")
        self.end_headers()
        if msg:
            self.wfile.write(msg.encode("utf-8"))
        else:
            self.wfile.write(b"Page not found")

    def _serve_project_listing(self, ct):
        projects = []
        with os.scandir(index_path) as it:
            for entry in it:
                if entry.is_dir():
                    projects.append({"name": entry.name})
        projects.sort(key=lambda proj: proj["name"])

        if ct in [PYPI_JSON_TYPE_V1, PYPI_JSON_TYPE_LT]:
            self.send_response(200)
            self.send_header("Content-Type", PYPI_JSON_TYPE_V1)
            self.end_headers()
            body = json.dumps({
                "meta": {"api-version": "1.0"},
                "projects": projects
            })
            self.wfile.write(body.encode("utf-8"))
        else:
            self.send_response(200)
            self.send_header("Content-Type", ct)
            self.end_headers()
            self.wfile.write(
                b"<!DOCTYPE html>\n<html>\n  <body>\n")
            for (i, project) in enumerate(projects):
                newline = "\n" if i < len(projects) - 1 else ""
                self.wfile.write(
                    "    <a href=\"/{}/\">{}</a>{}".format(
                        html.escape(project["name"]),
                        project["name"], newline).encode("utf-8"),
                )
            self.wfile.write(b"\n  </body>\n</html>")

    def _serve_project(self, ct, project):
        project = normalize(project)

        path = pathlib.Path(index_path, project)
        if not path.exists() or not path.is_dir():
            self._serve_notfound("No such project {}".format(project))
            return

        files = []
        with os.scandir(path) as it:
            for entry in it:
                if re.search(r"\.(whl|zip|tar\.gz)$", entry.name):
                    file = {
                        "filename": entry.name,
                        "url": "/{}/{}".format(project, entry.name),
                        "hashes": {},
                    }

                    # read file hash
                    hashfile = path.joinpath("{}.hash".format(entry.name))
                    if hashfile.exists():
                        with open(hashfile, "r") as hf:
                            data = hf.read().strip().split("=")
                            file["hashes"][data[0]] = data[1]

                    # do we have a metadata file?
                    if path.joinpath("{}.metadata".format(entry.name)):
                        file["dist-info-metadata"] = True

                    files.append(file)
        files.sort(key=lambda file: file["filename"])

        if ct in [PYPI_JSON_TYPE_V1, PYPI_JSON_TYPE_LT]:
            self.send_response(200)
            self.send_header("Content-Type", PYPI_JSON_TYPE_V1)
            self.end_headers()
            body = json.dumps({
                "name": project,
                "meta": {"api-version": "1.0"},
                "files": files
            })
            self.wfile.write(body.encode("utf-8"))
        else:
            self.send_response(200)
            self.send_header("Content-Type", ct)
            self.end_headers()
            self.wfile.write(
                b"<!DOCTYPE html>\n<html>\n  <body>\n")
            for (i, file) in enumerate(files):
                newline = "\n" if i < len(files) - 1 else ""
                hashval = ""
                if "sha256" in file["hashes"]:
                    hashval = "#{}={}".format(
                        "sha256", file["hashes"]["sha256"])
                self.wfile.write(
                    "    <a href=\"{}{}\" data-dist-info-metadata=\"{}\">{}</a>{}".format(
                        file["url"],
                        hashval,
                        "true" if file["dist-info-metadata"] else "false",
                        file["filename"],
                        newline,
                    ).encode("utf-8"),
                )
            self.wfile.write(b"\n  </body>\n</html>")

    def _serve_file(self, project, filename):
        project = normalize(project)

        path = pathlib.Path(index_path, project, filename)
        if not path.exists() or not path.is_file():
            self._serve_notfound("No such project {}".format(project))
            return

        ct = "text/plain"
        if re.search(r"\.(whl|zip)$", filename):
            ct = "application/octet-stream"
        elif re.search(r"\.tar\.gz$", filename):
            ct = "application/x-tar"

        self.send_response(200)
        self.send_header("Content-Type", ct)
        self.end_headers()
        self.wfile.write(path.read_bytes())


def run(ipath: str = os.getcwd(), host: str = "0.0.0.0", port: int = 8080):
    """
    Run the server on the provided index path, listening on the provided host
    and port. All arguments are options. By default, index path is the current
    working directory, and the server will listen on all interfaces at port
    8080.
    """

    global index_path
    index_path = ipath
    http.server.ThreadingHTTPServer(
        (host, port),
        RequestHandler,
    ).serve_forever()


def parse_accept_header(header_val: str) -> str:
    """
    Parses an Accept HTTP header and returns a selected MIME type for the server
    to answer with, honoring priorities defined in the header value. If the
    header value is empty, or the topmost priority is */*, HTML will be returned
    for backwards compatibility with PEP 503 clients.
    """

    if not header_val:
        return PYPI_HTML_TYPE_V1

    accepts_all = False

    options = [parse_accept_option(option) for option in header_val.split(",")]
    options.sort(key=lambda option: option["priority"], reverse=True)

    for option in options:
        if option["mime"] == "*/*":
            accepts_all = True

        if option["mime"] in [
            PYPI_JSON_TYPE_V1,
            PYPI_JSON_TYPE_LT,
            PYPI_HTML_TYPE_V1,
            GENL_HTML_TYPE
        ]:
            return option["mime"]

    if accepts_all:
        return PYPI_HTML_TYPE_V1

    return None


opt_re = re.compile(r"([^;]+)\s*(?:;.*q=(\d(?:\.\d+)?))")


def parse_accept_option(option: str) -> dict:
    """
    Parses one option from a multi-option Accept header, returning a dictionary
    with the MIME type (key "mime") and priority/quality (key "priority", a
    float). If option does not specifically list a priority, it will be zero.
    """

    m = opt_re.match(option)
    if m is None:
        return {"mime": option.strip(), "priority": 0}

    return {
        "mime": m.group(1).strip(),
        "priority": float(m.group(2)) if m.lastindex == 2 else 0
    }


def normalize(name):
    """
    Normalize the name of a package as per PEP 503.
    """

    return re.sub(r"[-_.]+", "-", name).lower()


def add_arguments(parser: argparse.ArgumentParser):
    """
    Used to add server-specific command line options.
    """

    parser.add_argument(
        '-H', '--host',
        dest='host',
        default='0.0.0.0',
        help='Host to listen on',
    )
    parser.add_argument(
        '-p', '--port',
        dest='port',
        default=8080,
        type=int,
        help='Port to listen on',
    )


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Morgan PyPI Server")
    parser.add_argument(
        '-i', '--index-path',
        dest='index_path',
        default=os.getcwd(),
        help='Path to the package index',
    )
    add_arguments(parser)
    args = parser.parse_args()

    run(args.index_path, args.host, args.port)
